import os, sys

from ..zincbase import KB

def get_cache_dir():
    path = os.path.abspath(sys.modules[KB.__module__].__file__)
    path = '/'.join(path.split('/')[:-1]) + '/.cache/'
    if not os.path.isdir(path):
        os.mkdir(path)
    return path

def check_file_exists(filename):
    return os.path.isfile(filename)


import urllib

import torch
import torch.nn as nn
from pytorch_pretrained_bert import BertModel

from zincbase.nn.tokenizer import get_tokenizer
from zincbase.utils.file_utils import get_cache_dir, check_file_exists
from zincbase.utils.string_utils import clean_punctuation

TOKEN_TYPES = ('<PAD>', 'O', 'B-LOC', 'I-LOC', 'B-PER', 'I-PER', 'B-ORG', 'I-ORG', 'B-MISC', 'I-MISC')
tag2idx = {tag: idx for idx, tag in enumerate(TOKEN_TYPES)}
idx2tag = {idx: tag for idx, tag in enumerate(TOKEN_TYPES)}

class BertNER(nn.Module):
    def __init__(self, vocab_size=None, device='cpu', training=False):
        super().__init__()
        self.bert = BertModel.from_pretrained('bert-base-cased').to(device)
        self.classifier = nn.Linear(768, vocab_size)
        self.device = device
        self.training = training
        self.bert.eval()

    def forward(self, x):
        x = x.to(self.device)
        if self.training:
            self.bert.train()
            layers_out, _ = self.bert(x)
            last_layer = layers_out[-1]
        else:
            with torch.no_grad():
                layers_out, _ = self.bert(x)
                last_layer = layers_out[-1]
        logits = self.classifier(last_layer)
        preds = logits.argmax(-1)
        return logits, preds

class NERModel():
    """Class for NER model.
    """
    def __init__(self, device='cpu', alternate_model_weights=None):
        """
        :param str device: 'cuda' or 'cpu', defaults to cpu.
        :param str alternate_model_weights: If you've trained your own model, specify the path to its .bin file here
        """
        self.device = device
        if not alternate_model_weights:
            self.model_name = 'ner_model.bin'
            weights_file = get_cache_dir() + self.model_name
        else:
            weights_file = alternate_model_weights
        
        if not check_file_exists(weights_file):
            print('Downloading weights file; afterwards it will be cached at %s' % weights_file)
            url = 'https://zincbase.com/models/ner_model.bin'
            urllib.request.urlretrieve(url, weights_file)

        self.ner_model = BertNER(len(TOKEN_TYPES), device, False).to(device)
        weights = torch.load(weights_file)
        self.ner_model.load_state_dict(weights)
    
    def ner(self, doc):
        """Carry out named entity recognition on doc, finding locations, people, organizations and misc things.

        :param str doc: A string of text, which might have named entities in it.
        :returns dict: Dictionary with keys LOC, PER, ORG and MISC, each of which is a list of found entities.

        :Example:

        >>> from nn.ner import NERModel
        >>> nlp = NERModel(device='cpu')
        >>> nlp.ner('The cat, Kitty, meowed in SOMA')
        {'LOC': ['SOMA'], 'PER': ['Kitty'], 'ORG': [], 'MISC': []}
        """

        tokens = []
        is_heads = []
        tmp_toks = []
        doc = '[CLS] ' + doc + ' [SEP]'
        for word in doc.split():
            tmp = get_tokenizer().tokenize(word.strip()) if word.strip() not in ("[CLS]", "[SEP]") else [word]
            tmp_toks.extend(tmp)
            xx = get_tokenizer().convert_tokens_to_ids(tmp)
            is_head = [1] + [0]*(len(xx) - 1)
            tokens.extend(xx)
            is_heads.extend(is_head)
        tokens = torch.LongTensor([tokens]).to(self.device)
        real_toks = []
        last_tok = ''
        for i, t in enumerate(tmp_toks):
            if is_heads[i]:
                if i != 0:
                    real_toks.append(last_tok)
                last_tok = t
            else:
                last_tok += t.replace('##', '')
        real_toks.append(last_tok)
        _, toktype = self.ner_model(tokens)
        toktype = toktype.cpu().numpy().tolist()
        y_pred = [p for head, p in zip(is_heads, toktype[0]) if head == 1]
        preds = [idx2tag[p] for p in y_pred]
        ents = {
            'LOC': [],
            'PER': [],
            'ORG': [],
            'MISC': [],
        }
        cat = ''
        enttype = ''
        for pred, real_tok in list(zip(preds, real_toks)):
            if pred == 'O':
                if cat and enttype:
                    ents[enttype].append(clean_punctuation(cat))
                cat = ''
                continue
            if pred[0] == 'B':
                cat = real_tok
                enttype = pred.split('-')[-1]
            if pred[0] == 'I':
                cat += ' ' + real_tok
        if cat:
            ents[enttype].append(clean_punctuation(cat))
        return ents


"""A base unit for ZincBase's Prolog-like implementation of 'facts'"""

from zincbase.utils.string_utils import split_on

class Term:
    def __init__(self, expr, args=None, graph=None):
        if args:
            self.pred = expr
            self.args = args
        elif expr[-1] == ']':
            arr = split_on(expr[1:-1], ',')
            headtail = split_on(expr[1:-1], '|')
            if len(headtail) > 1:
                self.args = [Term(f, graph=graph) for f in headtail]
                self.pred = '__list__'
            else:
                arr.reverse()
                first = Term('__list__', [], graph=graph)
                for part in arr:
                    first = Term('__list__', [Term(part, graph=graph), first], graph=graph)
                self.pred = first.pred
                self.args = first.args
        elif expr[-1] == ')':
            sub_exprs = split_on(expr, '(', all=False)
            if len(sub_exprs) != 2:
                raise Exception('Syntax error')
            self.args = [Term(sub_expr, graph=graph) for sub_expr in split_on(sub_exprs[1][:-1], ',')]
            self.pred = sub_exprs[0]
        else:
            self.pred = expr
            self.args = []

        if graph is not None:
            for i, arg in enumerate(self.args):
                if arg:
                    if not graph.has_node(str(arg)):
                        graph.add_node(str(arg))
                    for arg2 in self.args[i+1:]:
                        if not graph.has_node(str(arg2)):
                            graph.add_node(str(arg2))
                        graph.add_edge(str(arg), str(arg2), pred=self.pred)

    def __repr__(self):
        if self.pred == '__list__':
            if not self.args:
                return '[]'
            first = self.args[1]
            if first.pred == '__list__' and first.args == []:
                return '[{}]'.format(str(self.args[0]))
            elif first.pred == '__list__':
                return '[{},{}]'.format(str(self.args[0]), str(self.args[1])[1:-1])
            else:
                return '[{}|{}]'.format(str(self.args[0]), str(self.args[1]))
        elif self.args:
            return '{}({})'.format(self.pred, ', '.join(map(str,self.args)))
        else:
            return self.pred



from pytorch_pretrained_bert import BertTokenizer

tokenizer = None

def get_tokenizer(bert_model='bert-base-cased', do_lower_case=False):
    global tokenizer
    if tokenizer:
        return tokenizer
    tokenizer = BertTokenizer.from_pretrained(bert_model, do_lower_case=do_lower_case)
    return tokenizer




from distutils.core import setup
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    requirements = fh.read()

setup(
    name='Zincbase',
    version='0.2.61',
    packages=setuptools.find_packages(),
    description="A state of the art knowledge base and batteries-included NLP toolkit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    include_package_data=True,
    url='https://github.com/tomgrek/zincbase',
    author='Tom Grek',
    author_email='tom.grek@gmail.com',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


import context
from zincbase import KB

kb = KB()
kb.store('a(b,c)')
kb.attr('b', {'is_letter': 1.0})
assert kb.node('b') == {'is_letter': 1.0}
kb.edge_attr('b', 'a', 'c', {'both_alpha': 1.0})
assert kb.edge('b', 'a', 'c') == {'both_alpha': 1.0}
assert kb.to_triples() == [('b', 'a', 'c')]
triples = kb.to_triples(data=True)
assert triples == [('b', 'a', 'c', {'is_letter': 1.0}, {'both_alpha': 1.0}, {}, False)]
kb.attr('c', {'is_letter': 0.9})
triples = kb.to_triples(data=True)
assert triples == [('b', 'a', 'c', {'is_letter': 1.0}, {'both_alpha': 1.0}, {'is_letter': 0.9}, False)]
neg_rule_idx = kb.store('~a(b,c)')
triples = kb.to_triples(data=True)
assert triples == [('b', 'a', 'c', {'is_letter': 1.0}, {'both_alpha': 1.0}, {'is_letter': 0.9}, True)]
kb.delete_rule(neg_rule_idx)
triples = kb.to_triples(data=True)
assert triples == [('b', 'a', 'c', {'is_letter': 1.0}, {'both_alpha': 1.0}, {'is_letter': 0.9}, False)]
kb.edge_attr('b', 'a', 'c', {'truthiness':-1})
triples = kb.to_triples(data=True)
assert triples == [('b', 'a', 'c', {'is_letter': 1.0}, {'both_alpha': 1.0, 'truthiness': -1}, {'is_letter': 0.9}, True)]
kb.delete_edge_attr('b', 'a', 'c', ['truthiness'])
triples = kb.to_triples(data=True)
assert triples == [('b', 'a', 'c', {'is_letter': 1.0}, {'both_alpha': 1.0}, {'is_letter': 0.9}, False)]
print('All graph tests passed.')